# File autogenerated during install.
# Do not change. Do not store in version control.

__LZ4_VERSION = '1.8.0'
@deprecated('use lz4.lz4version() instead')
def LZ4_VERSION_f():
    return __LZ4_VERSION

LZ4_VERSION = LZ4_version_f()
